from functools import wraps
import inspect
import warnings

# DeprecationWarning is ignored by default, so we define a subcategory that we can give a different default.


class ManticoreDeprecationWarning(DeprecationWarning):
    """The deprecation warning class used by Manticore."""

    pass


warnings.simplefilter("default", category=ManticoreDeprecationWarning)


def deprecated(message: str):
    """A decorator for marking functions as deprecated. """
    assert isinstance(message, str), "The deprecated decorator requires a message string argument."

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            warnings.warn(
                f"`{func.__qualname__}` is deprecated. {message}",
                category=ManticoreDeprecationWarning,
                stacklevel=2,
            )
            return func(*args, **kwargs)

        return wrapper

    return decorator



from .aarch64 import Aarch64Cpu, Aarch64CdeclAbi, Aarch64LinuxSyscallAbi
from .arm import Armv7Cpu, Armv7CdeclAbi, Armv7LinuxSyscallAbi
from .x86 import (
    AMD64Cpu,
    I386Cpu,
    AMD64LinuxSyscallAbi,
    I386LinuxSyscallAbi,
    I386CdeclAbi,
    SystemVAbi,
)
from .abstractcpu import Abi, Cpu, SyscallAbi


class CpuFactory:
    _cpus = {"i386": I386Cpu, "amd64": AMD64Cpu, "armv7": Armv7Cpu, "aarch64": Aarch64Cpu}

    _linux_abis = {
        "i386": I386CdeclAbi,
        "amd64": SystemVAbi,
        "armv7": Armv7CdeclAbi,
        "aarch64": Aarch64CdeclAbi,
    }

    _linux_syscalls_abis = {
        "i386": I386LinuxSyscallAbi,
        "amd64": AMD64LinuxSyscallAbi,
        "armv7": Armv7LinuxSyscallAbi,
        "aarch64": Aarch64LinuxSyscallAbi,
    }

    @staticmethod
    def get_cpu(mem, machine: str) -> Cpu:
        cpu = CpuFactory._cpus[machine](mem)
        mem.cpu = cpu
        return cpu

    @staticmethod
    def get_function_abi(cpu: Cpu, os: str, machine: str) -> Abi:
        if os != "linux" or machine not in CpuFactory._linux_abis:
            raise NotImplementedError(f"OS and machine combination not supported: {os}/{machine}")

        return CpuFactory._linux_abis[machine](cpu)

    @staticmethod
    def get_syscall_abi(cpu: Cpu, os: str, machine: str) -> SyscallAbi:
        if os != "linux" or machine not in CpuFactory._linux_syscalls_abis:
            raise NotImplementedError(f"OS and machine combination not supported: {os}/{machine}")

        return CpuFactory._linux_syscalls_abis[machine](cpu)



"""
Models here are intended to be passed to :meth:`~manticore.native.state.State.invoke_model`, not invoked directly.
"""

from .cpu.abstractcpu import Cpu, ConcretizeArgument
from .state import State
from ..core.smtlib import issymbolic, BitVec
from ..core.smtlib.operators import ITEBV, ZEXTEND
from ..core.state import Concretize
from typing import Union

VARIADIC_FUNC_ATTR = "_variadic"


def isvariadic(model):
    """
    :param callable model: Function model
    :return: Whether `model` models a variadic function
    :rtype: bool
    """
    return getattr(model, VARIADIC_FUNC_ATTR, False)


def variadic(func):
    """
    A decorator used to mark a function model as variadic. This function should
    take two parameters: a :class:`~manticore.native.state.State` object, and
    a generator object for the arguments.

    :param callable func: Function model
    """
    setattr(func, VARIADIC_FUNC_ATTR, True)
    return func


def must_be_NULL(state, byte) -> bool:
    """
    Checks if a given byte read from memory is NULL.
    This supports both concrete & symbolic byte values.

    :param byte: byte read from memory to be examined
    :param constrs: state constraints
    :return: whether a given byte is NULL or constrained to NULL
    """
    if issymbolic(byte):
        return state.must_be_true(byte == 0)
    else:
        return byte == 0


def cannot_be_NULL(state, byte) -> bool:
    """
    Checks if a given byte read from memory is not NULL or cannot be NULL

    :param byte: byte read from memory to be examined
    :param constrs: state constraints
    :return: whether a given byte is not NULL or cannot be NULL
    """
    if issymbolic(byte):
        return state.must_be_true(byte != 0)
    else:
        return byte != 0


def can_be_NULL(state, byte) -> bool:
    """
    Checks if a given byte read from memory can be NULL

    :param byte: byte read from memory to be examined
    :param constrs: state constraints
    :return: whether a given byte is NULL or can be NULL
    """
    if issymbolic(byte):
        return state.can_be_true(byte == 0)
    else:
        return byte == 0


def _find_zero(cpu, state, ptr: Union[int, BitVec]) -> int:
    """
    Helper for finding the closest NULL or, effectively NULL byte from a starting address.

    :param Cpu cpu:
    :param ConstraintSet constrs: Constraints for current `State`
    :param int ptr: Address to start searching for a zero from
    :return: Offset from `ptr` to first byte that is 0 or an `Expression` that must be zero
    """

    offset = 0
    while True:
        byte = cpu.read_int(ptr + offset, 8)
        if must_be_NULL(state, byte):
            break
        offset += 1

    return offset


def strcmp(state: State, s1: Union[int, BitVec], s2: Union[int, BitVec]):
    """
    strcmp symbolic model.

    Algorithm: Walks from end of string (minimum offset to NULL in either string)
    to beginning building tree of ITEs each time either of the
    bytes at current offset is symbolic.

    Points of Interest:
    - We've been building up a symbolic tree but then encounter two
    concrete bytes that differ. We can throw away the entire symbolic
    tree!
    - If we've been encountering concrete bytes that match
    at the end of the string as we walk forward, and then we encounter
    a pair where one is symbolic, we can forget about that 0 `ret` we've
    been tracking and just replace it with the symbolic subtraction of
    the two

    :param state: Current program state
    :param s1: Address of string 1
    :param s2: Address of string 2
    :return: Symbolic strcmp result
    :rtype: Expression or int
    """

    cpu = state.cpu

    if issymbolic(s1):
        raise ConcretizeArgument(state.cpu, 1)
    if issymbolic(s2):
        raise ConcretizeArgument(state.cpu, 2)

    s1_zero_idx = _find_zero(cpu, state, s1)
    s2_zero_idx = _find_zero(cpu, state, s2)
    min_zero_idx = min(s1_zero_idx, s2_zero_idx)

    ret = None

    for offset in range(min_zero_idx, -1, -1):
        s1char = ZEXTEND(cpu.read_int(s1 + offset, 8), cpu.address_bit_size)
        s2char = ZEXTEND(cpu.read_int(s2 + offset, 8), cpu.address_bit_size)

        if issymbolic(s1char) or issymbolic(s2char):
            if ret is None or (not issymbolic(ret) and ret == 0):
                ret = s1char - s2char
            else:
                ret = ITEBV(cpu.address_bit_size, s1char != s2char, s1char - s2char, ret)
        else:
            if s1char != s2char:
                ret = s1char - s2char
            elif ret is None:
                ret = 0

    return ret


def strlen_exact(state: State, s: Union[int, BitVec]) -> Union[int, BitVec]:
    """
    strlen symbolic model

    Strategy: produce a state for every symbolic string length for better accuracy

    Algorithm: Counts the number of characters in a string forking every time a symbolic byte
    is found that can be NULL but is not constrained to NULL.

    :param state: current program state
    :param s: Address of string
    :return: Symbolic strlen result
    """

    if issymbolic(s):
        raise ConcretizeArgument(state.cpu, 1)

    cpu = state.cpu

    # Initialize offset based on whether state has been forked in strlen
    if "strlen" not in state.context:
        offset = 0
    else:
        offset = state.context["strlen"]

    c = cpu.read_int(s + offset, 8)
    while not must_be_NULL(state, c):
        # If the byte can be NULL concretize and fork states
        if can_be_NULL(state, c):
            state.context["strlen"] = offset
            raise Concretize("Forking on possible NULL strlen", expression=(c == 0), policy="ALL")

        offset += 1
        c = cpu.read_int(s + offset, 8)

    return offset


def strlen_approx(state: State, s: Union[int, BitVec]) -> Union[int, BitVec]:
    """
    strlen symbolic model

    Strategy: build a result tree to limit state explosion results approximate

    Algorithm: Walks from end of string not including NULL building ITE tree when current byte is symbolic.

    :param state: current program state
    :param s: Address of string
    :return: Symbolic strlen result
    """

    if issymbolic(s):
        raise ConcretizeArgument(state.cpu, 1)

    cpu = state.cpu
    zero_idx = _find_zero(cpu, state, s)

    ret = zero_idx

    for offset in range(zero_idx - 1, -1, -1):
        byt = cpu.read_int(s + offset, 8)
        if issymbolic(byt):
            ret = ITEBV(cpu.address_bit_size, byt == 0, offset, ret)

    return ret


def strcpy(state: State, dst: Union[int, BitVec], src: Union[int, BitVec]) -> Union[int, BitVec]:
    """
    strcpy symbolic model

    Algorithm: Copy every byte from src to dst until finding a byte that is NULL or is
    constrained to only the NULL value. Every time a byte is fouund that can be NULL but
    is not definetly NULL concretize and fork states.

    :param state: current program state
    :param dst: destination string address
    :param src: source string address
    :return: pointer to the dst
    """
    if issymbolic(src):
        raise ConcretizeArgument(state.cpu, 2)

    if issymbolic(dst):
        raise ConcretizeArgument(state.cpu, 1)

    cpu = state.cpu
    ret = dst

    # Initialize offset based on whether state has been forked in strcpy
    if "strcpy" not in state.context:
        offset = 0
    else:
        offset = state.context["strcpy"]

    # Copy until a src_byte is symbolic and constrained to '\000', or is concrete and '\000'
    src_val = cpu.read_int(src + offset, 8)
    while not must_be_NULL(state, src_val):
        cpu.write_int(dst + offset, src_val, 8)

        # If a src byte can be NULL concretize and fork states
        if can_be_NULL(state, src_val):
            state.context["strcpy"] = offset
            raise Concretize("Forking on NULL strcpy", expression=(src_val == 0), policy="ALL")
        offset += 1

        src_val = cpu.read_int(src + offset, 8)

    # Write concrete null for end of string in current state
    cpu.write_int(dst + offset, 0, 8)

    if "strcpy" in state.context:
        del state.context["strcpy"]
    return ret


def strncpy(
    state: State, dst: Union[int, BitVec], src: Union[int, BitVec], n: Union[int, BitVec]
) -> Union[int, BitVec]:
    """
    strncpy symbolic model

    Algorithm:  Copy n bytes from src to dst. If the length of the src string is less than n pad the difference
    with NULL bytes. If a symbolic byte is found that can be NULL but is not definitely NULL fork and concretize states.

    :param state: current program state
    :param dst: destination string address
    :param src: source string address
    :param n: number of bytes to copy
    :return: pointer to the dst
    """

    if issymbolic(dst):
        raise ConcretizeArgument(state.cpu, 1)
    if issymbolic(src):
        raise ConcretizeArgument(state.cpu, 2)
    if issymbolic(n):
        raise ConcretizeArgument(state.cpu, 3)

    cpu = state.cpu
    ret = dst

    # Initialize offset based on whether state has been forked in strncpy
    if "strncpy" not in state.context:
        offset = 0
    else:
        offset = state.context["strncpy"]

    # Copy until a src_byte is symbolic and constrained to '\000', or is concrete and '\000'
    src_val = cpu.read_int(src + offset, 8)
    while offset < n and not must_be_NULL(state, src_val):
        cpu.write_int(dst + offset, src_val, 8)

        # If a src byte can be NULL concretize and fork states
        if can_be_NULL(state, src_val):
            state.context["strncpy"] = offset
            raise Concretize("Forking on NULL strncpy", expression=(src_val == 0), policy="ALL")
        offset += 1

        src_val = cpu.read_int(src + offset, 8)

    # Pad the distance between length of src and n with NULL bytes
    while offset < n:
        cpu.write_int(dst + offset, 0, 8)
        offset += 1

    if "strncpy" in state.context:
        del state.context["strncpy"]
    return ret



from abc import abstractmethod

import capstone as cs


class Instruction:
    """Capstone-like instruction to be used internally"""

    @property
    @abstractmethod
    def address(self):
        pass

    @property
    @abstractmethod
    def mnemonic(self):
        pass

    @property
    @abstractmethod
    def op_str(self):
        pass

    @property
    @abstractmethod
    def size(self):
        pass

    @property
    @abstractmethod
    def operands(self):
        pass

    # FIXME (theo) eliminate one of the two of insn_name, name
    @property
    @abstractmethod
    def insn_name(self):
        pass

    @property
    @abstractmethod
    def name(self):
        pass


class Disasm:
    """Abstract class for different disassembler interfaces"""

    def __init__(self, disasm):
        self.disasm = disasm

    @abstractmethod
    def disassemble_instruction(self, code, pc):
        """Get next instruction based on the disassembler in use

        :param str code: binary blob to be disassembled
        :param long pc: program counter
        """


class CapstoneDisasm(Disasm):
    def __init__(self, arch, mode):
        try:
            cap = cs.Cs(arch, mode)
        except Exception as e:
            raise e
        cap.detail = True
        cap.syntax = 0
        super().__init__(cap)

    def disassemble_instruction(self, code, pc):
        """Get next instruction using the Capstone disassembler

        :param str code: binary blob to be disassembled
        :param long pc: program counter
        """
        return next(self.disasm.disasm(code, pc))


def init_disassembler(disassembler, arch, mode, view=None):
    if disassembler == "capstone":
        return CapstoneDisasm(arch, mode)
    else:
        raise NotImplementedError("Disassembler not implemented")



import collections
import logging
import pickle
import string
import sys
from collections import OrderedDict
from gzip import GzipFile
from io import BytesIO
from prettytable import PrettyTable
from datetime import datetime

from typing import Any, IO, Dict

from .config import get_group
from .enums import StateLists

logger = logging.getLogger(__name__)

consts = get_group("core")
consts.add(
    "compress_states",
    default=True,
    description="Seamlessly compress state files on disk. Reduces space usage and improves performance on slow disks, "
    "at the cost of some slight [de]compression overhead.",
)


def interval_intersection(min1, max1, min2, max2):
    """
    Given two intervals, (min1, max1) and (min2, max2) return their intersecting interval,
    or None if they do not overlap.
    """
    left, right = max(min1, min2), min(max1, max2)
    if left < right:
        return left, right
    return None


def printable_bytes(bytes: bytes):
    return "".join([c for c in map(chr, bytes) if c in string.printable])


class CacheDict(OrderedDict):
    def __init__(self, *args, max_size=30000, flush_perc=30, **kwargs):
        self._max_size = max_size
        self._purge_percent = flush_perc * 0.01
        self._misses = 0
        self._hits = 0
        self._flushes = 0
        super().__init__(*args, **kwargs)

    def __del__(self):
        try:
            log = logging.getLogger(self.__class__.__name__)
            log.debug(
                f"DictCache: hits: {self._hits}, misses: {self._misses}, flushes: {self._flushes}, size: {self.__len__()}"
            )
        except TypeError:
            # Prevent "TypeError: attribute of type 'NoneType' is not callable" on line 32
            # TODO - figure out why this happens (I think it's only on concrete runs?)
            pass

    def __setitem__(self, key, value):
        if len(self) > self._max_size:
            self.flush()
        return super().__setitem__(key, value)

    def __contains__(self, item):
        x = super().__contains__(item)
        if x:
            self._hits += 1
        else:
            self._misses += 1
        return x

    def flush(self):
        self._flushes += 1
        purge_count = int(len(self) * self._purge_percent)
        for i in range(purge_count):
            self.popitem(last=False)
        self._hits -= purge_count


class StateSerializer:
    """
    StateSerializer can serialize and deserialize :class:`~manticore.core.state.State` objects from and to
    stream-like objects.
    """

    def __init__(self):
        pass

    def serialize(self, state, f):
        raise NotImplementedError

    def deserialize(self, f):
        raise NotImplementedError


class PickleSerializer(StateSerializer):
    """
    A StateSerializer that uses a gzip-based Python pickle format.
    """

    DEFAULT_RECURSION: int = 0x10000  # 1M
    MAX_RECURSION: int = 0x1000000  # 16.7M
    COMPRESSION_LEVEL: int = 1  # minimal compression, but still gets >10x reduction

    def __init__(self):
        super().__init__()
        sys.setrecursionlimit(PickleSerializer.DEFAULT_RECURSION)

    def serialize(self, state, f):
        logger.info("Serializing %s", f.name if hasattr(f, "name") else "<unknown>")
        try:
            pickle_dump(
                state,
                GzipFile(fileobj=f, mode="wb", compresslevel=PickleSerializer.COMPRESSION_LEVEL)
                if consts.compress_states
                else f,
            )
        except RuntimeError:
            new_limit = sys.getrecursionlimit() * 2
            if new_limit > PickleSerializer.MAX_RECURSION:
                raise Exception(
                    f"PickleSerializer recursion limit surpassed {PickleSerializer.MAX_RECURSION}, aborting"
                )
            logger.info(f"Recursion soft limit {sys.getrecursionlimit()} hit, increasing")
            sys.setrecursionlimit(new_limit)
            self.serialize(state, f)

    def deserialize(self, f):
        logger.info("Deserializing %s", f.name if hasattr(f, "name") else "<unknown>")
        return pickle.load(GzipFile(fileobj=f, mode="rb") if consts.compress_states else f)


def pickle_dumps(obj: Any) -> bytes:
    """
    Serializes an object as a gzipped pickle.
    """
    # This consolidates pickling in one place so we can fix the protocol version
    fp = BytesIO()
    pickle_dump(obj, fp)
    return fp.getvalue()


def pickle_dump(obj: Any, fp: IO[bytes]) -> None:
    """
    Serializes an object as a gzipped pickle to the given file.
    """
    return pickle.dump(obj, fp, protocol=pickle.HIGHEST_PROTOCOL)


def pretty_print_state_descriptors(desc: Dict):
    """
    Given a dict of state descriptors, nicely formats and prints it to stdout.
    :param desc: Dict mapping state IDs to State Descriptors, like the one returned from MantioreBase.introspect
    """

    descriptors = desc.values()
    nready, nbusy, nkill, nterm = 0, 0, 0, 0
    for st in descriptors:
        nready += 1 if (st.state_list == StateLists.ready) else 0
        nbusy += 1 if (st.state_list == StateLists.busy) else 0
        nkill += 1 if (st.state_list == StateLists.busy) else 0
        nterm += 1 if (st.state_list == StateLists.terminated) else 0

    print(
        "Ready States:",
        nready,
        " | ",
        "Busy States:",
        nbusy,
        " | ",
        "Terminated States:",
        nterm,
        " | ",
        "Killed States:",
        nkill,
    )

    tab = PrettyTable()
    tab.field_names = ["ID", "Status", "Duration", "Execs", "Execs/Sec"]
    if nbusy:
        now = datetime.now()
        for st in descriptors:
            if st.state_list == StateLists.busy:
                duration = (
                    now - st.field_updated_at["state_list"]
                )  # Time since this state became Busy
                execs = st.own_execs if st.own_execs is not None else 0
                tab.add_row(
                    [
                        st.state_id,
                        st.status.value,
                        str(duration)[:-4],
                        execs,
                        "{:.2f}".format(execs / (now - st.created_at).total_seconds()),
                    ]
                )

        print(tab)
    print()


class deque(collections.deque):
    """ A wrapper around collections.deque that adds a few APIs present in SyncManager.Queue """

    def empty(self) -> bool:
        return len(self) == 0

    def get(self):
        return self.popleft()



""" Common binary formats interface
Ideally you should be able to do something like

        from binary import Binary
        binary = Binary(filename)
        assert cpu.machine == binary.arch, "Not matching cpu"
        logger.info(f"Loading {filename} as a {binary.arch} elf")
        for mm in binary.maps():
            cpu.mem.mmapFile( mm )
        for th in binary.threads():
            setup(th)

But there are difference between format that makes it difficult to find a simple
and common API.  interpreters? linkers? linked DLLs?

"""

from .binary import Binary, CGCElf, Elf  # noqa


if __name__ == "__main__":
    import sys

    print(list(Binary(sys.argv[1]).threads()))
    print(list(Binary(sys.argv[1]).maps()))



from functools import wraps
from inspect import signature as inspect_signature
import logging
import struct

import capstone as cs
import operator as ops

from .abstractcpu import Abi, Cpu, Interruption, Operand, RegisterFile, SyscallAbi
from .abstractcpu import instruction as abstract_instruction
from .bitwise import *
from .register import Register
from ...core.smtlib import Operators, BitVecConstant, issymbolic

from typing import NamedTuple

logger = logging.getLogger(__name__)

# map different instructions to a single impl here
OP_NAME_MAP = {"MOVW": "MOV"}


def HighBit(n):
    return Bit(n, 31)


ARMV7_CPU_ADDRESS_BIT_SIZE = 32


def instruction(instruction_body=None, *, can_take_denormalized_mod_imm: bool = False):
    """
    This decorator is used to annotate Armv7Cpu methods as
    instruction-implementing methods.

    This centralizes some common ARM-specific logic about CPU flags in one place.

    Additionally, this optionally adds /modified immediate normalization/ logic
    to the wrapped method.

    This decorator works both as `@instruction` and as
    `@instruction(can_take_denormalized_mod_imm=True)`.


    What is this normalization logic?

    First, it helps to understand how ARM encodes immediate constants.
    In encoded ARM instructions, immediate constant values are encoded as an
    8-bit unsigned number and a 4-bit rotation value; you can read about the
    details in the ARM Architecture Reference Manual, ARMv7-A and ARMv7-R
    edition, section A5.2.3, "Data-processing (immediate)".

    Second, it turns out that the Capstone disassembler we use will sometimes
    disassemble an ARM immediate constant value into /two/ immediate operand
    values, explicitly representing the 8-bit unsigned number and two times the
    4-bit shift.  In particular, it seems that Capstone uses this explicit
    representation when the modified immediate value is encoded in a
    non-canonical form.  A blog post has some more explanation around this:

        https://alisdair.mcdiarmid.org/arm-immediate-value-encoding/

    So, finally, the /modified immediate normalization/ logic that this
    decorator adds converts an explicitly-represented unsigned number and
    rotation into a single immediate operand-like value (`_ImmediatePseudoOperand`)
    that has the appropriate integer value, so that the actual implementation
    of an ARM instruction here can expect only normalized immediates, and not
    have to concern itself with this quirk of Capstone.
    """

    def decorator(body):
        if can_take_denormalized_mod_imm:
            # Need to possibly normalize a modified immediate argument that's
            # been explicitly split by Capstone into (number, rotation)
            # components.

            body_sig = inspect_signature(body)
            # subtract 1 to account for the first parameter (`cpu`), present in
            # all instruction methods.
            num_body_params = len(body_sig.parameters) - 1
            assert num_body_params > 0

            def normalize_mod_imm_arg(args):
                if len(args) == num_body_params + 1:
                    # We got 1 more argument than the wrapped function expects;
                    # this is the case of a modified immediate represented
                    # explicitly as 2 immediate operands.
                    # Normalize the two into one!
                    args = list(args)
                    rot = args.pop()
                    assert rot.type == "immediate"
                    num = args.pop()
                    assert num.type == "immediate"
                    imm = ROR(num.imm, rot.imm, ARMV7_CPU_ADDRESS_BIT_SIZE)
                    args.append(_ImmediatePseudoOperand(imm))
                return args

        else:
            # No normalization of a modified immediate required; return what's given.
            def normalize_mod_imm_arg(args):
                return args

        @wraps(body)
        def instruction_implementation(cpu, *args, **kwargs):
            should_execute = cpu.should_execute_conditional()

            if cpu._at_symbolic_conditional == cpu.instruction.address:
                cpu._at_symbolic_conditional = None
                should_execute = True
            else:
                if issymbolic(should_execute):
                    # Let's remember next time we get here we should not do this again
                    cpu._at_symbolic_conditional = cpu.instruction.address
                    i_size = cpu.instruction.size
                    cpu.PC = Operators.ITEBV(
                        cpu.address_bit_size, should_execute, cpu.PC - i_size, cpu.PC
                    )
                    return

            if should_execute:
                ret = body(cpu, *normalize_mod_imm_arg(args), **kwargs)
            else:
                ret = None

            if cpu.should_commit_flags():
                cpu.commit_flags()

            return ret

        return abstract_instruction(instruction_implementation)

    # Here's where we support using this decorator both like
    # `@instruction` and like `@instruction(can_take_denormalized_mod_imm=True)`.
    # See https://stackoverflow.com/questions/3888158/making-decorators-with-optional-arguments
    # for some decorator-fu.
    if instruction_body is not None:
        return decorator(instruction_body)
    else:
        return decorator


class _ImmediatePseudoOperand(NamedTuple):
    """
    This is a hacky class that is used to represent an object that looks close
    enough to an Armv7Operand to be used in the places where immediate operands
    are used.

    See the `instruction` decorator for more detail.
    """

    imm: int

    def read(self, with_carry: bool = False):
        if with_carry:
            return self.imm, 0
        else:
            return self.imm


_TYPE_MAP = {
    cs.arm.ARM_OP_REG: "register",
    cs.arm.ARM_OP_MEM: "memory",
    cs.arm.ARM_OP_IMM: "immediate",
    cs.arm.ARM_OP_PIMM: "coprocessor",
    cs.arm.ARM_OP_CIMM: "immediate",
}


class Armv7Operand(Operand):
    def __init__(self, cpu, op, **kwargs):
        super().__init__(cpu, op, **kwargs)
        self.__type = _TYPE_MAP[self.op.type]

    @property
    def type(self):
        """
        Corresponds to capstone's `operand.type` (cs.arm.ARM_OP_*).
        """
        return self.__type

    @property
    def size(self):
        assert self.__type == "register"
        if cs.arm.ARM_REG_D0 <= self.op.reg <= cs.arm.ARM_REG_D31:
            return 64
        else:
            # FIXME check other types of operand sizes
            return 32

    def read(self, nbits=None, with_carry=False):
        carry = self.cpu.regfile.read("APSR_C")
        if self.__type == "register":
            value = self.cpu.regfile.read(self.reg)
            # PC in this case has to be set to the instruction after next. PC at this point
            # is already pointing to next instruction; we bump it one more.
            if self.reg in ("PC", "R15"):
                value += self.cpu.instruction.size
            if self.is_shifted():
                shift = self.op.shift
                # XXX: This is unnecessary repetition.
                if shift.type in range(cs.arm.ARM_SFT_ASR_REG, cs.arm.ARM_SFT_RRX_REG + 1):
                    if self.cpu.mode == cs.CS_MODE_THUMB:
                        amount = shift.value.read()
                    else:
                        src_reg = self.cpu.instruction.reg_name(shift.value).upper()
                        amount = self.cpu.regfile.read(src_reg)
                else:
                    amount = shift.value
                value, carry = self.cpu._shift(value, shift.type, amount, carry)
            if self.op.subtracted:
                value = -value
            if with_carry:
                return value, carry
            return value
        elif self.__type == "immediate":
            imm = self.op.imm
            if self.op.subtracted:
                imm = -imm
            if with_carry:
                return imm, self._get_expand_imm_carry(carry)
            return imm
        elif self.__type == "coprocessor":
            imm = self.op.imm
            return imm
        elif self.__type == "memory":
            val = self.cpu.read_int(self.address(), nbits)
            if with_carry:
                return val, carry
            return val
        else:
            raise NotImplementedError("readOperand unknown type", self.op.type)

    def write(self, value, nbits=None):
        if self.__type == "register":
            self.cpu.regfile.write(self.reg, value)
        elif self.__type == "memory":
            raise NotImplementedError("need to impl arm store mem")
        else:
            raise NotImplementedError("writeOperand unknown type", self.op.type)

    def writeback(self, value):
        if self.__type == "register":
            self.write(value)
        elif self.__type == "memory":
            self.cpu.regfile.write(self.mem.base, value)
        else:
            raise NotImplementedError("writeback Operand unknown type", self.op.type)

    def is_shifted(self):
        """
        In ARM some of the operands may have an additional metadata which means they can be shifted
        with either a register or immediate value.

        See:
        * https://github.com/aquynh/capstone/blob/fdebc371ba0568acde007e08dad2cc3c9333e3fa/include/arm.h#L22-L34

        * 11.5 Syntax of Operand2 as a register with optional shift
            http://infocenter.arm.com/help/index.jsp?topic=/com.arm.doc.dui0473m/dom1361289852638.html

        :return: True if operand is shifted, otherwise False.
        """
        return self.op.shift.type != cs.arm.ARM_SFT_INVALID

    def address(self):
        assert self.__type == "memory"
        addr = self.get_mem_base_addr() + self.get_mem_offset()
        return addr & Mask(self.cpu.address_bit_size)

    def get_mem_offset(self):
        assert self.__type == "memory"

        off = 0
        if self.mem.index is not None:
            idx = self.mem.scale * self.cpu.regfile.read(self.mem.index)
            carry = self.cpu.regfile.read("APSR_C")
            if self.is_shifted():
                shift = self.op.shift
                idx, carry = self.cpu._shift(idx, shift.type, shift.value, carry)
            off = -idx if self.op.subtracted else idx
        else:
            off = self.mem.disp
        return off

    def get_mem_base_addr(self):
        assert self.__type == "memory"

        base = self.cpu.regfile.read(self.mem.base)

        # PC relative addressing is fun in ARM:
        # In ARM mode, the spec defines the base value as current insn + 8
        # In thumb mode, the spec defines the base value as ALIGN(current insn address) + 4,
        # where ALIGN(current insn address) => <current insn address> & 0xFFFFFFFC
        #
        # Regardless of mode, our implementation of read(PC) will return the address
        # of the instruction following the next instruction.
        if self.mem.base in ("PC", "R15"):
            if self.cpu.mode == cs.CS_MODE_ARM:
                logger.debug(f"ARM mode PC relative addressing: PC + offset: 0x{base:x} + 0x{4:x}")
                return base + 4
            else:
                # base currently has the value PC + len(current_instruction)
                # we need (PC & 0xFFFFFFFC) + 4
                # thus:
                new_base = (base - self.cpu.instruction.size) & 0xFFFFFFFC
                logger.debug(
                    f"THUMB mode PC relative addressing: ALIGN(PC) + offset => 0x{new_base:x} + 0x{4:x}"
                )
                return new_base + 4
        else:
            return base

    def _get_expand_imm_carry(self, carryIn):
        """Manually compute the carry bit produced by expanding an immediate operand (see ARMExpandImm_C)"""
        insn = struct.unpack("<I", self.cpu.instruction.bytes)[0]
        unrotated = insn & Mask(8)
        shift = Operators.EXTRACT(insn, 8, 4)
        _, carry = self.cpu._shift(unrotated, cs.arm.ARM_SFT_ROR, 2 * shift, carryIn)
        return carry


class Armv7RegisterFile(RegisterFile):
    def __init__(self):
        """
        ARM Register file abstraction. GPRs use ints for read/write. APSR
        flags allow writes of bool/{1, 0} but always read bools.
        """
        super().__init__(
            {
                "SB": "R9",
                "SL": "R10",
                "FP": "R11",
                "IP": "R12",
                "STACK": "R13",
                "SP": "R13",
                "LR": "R14",
                "PC": "R15",
            }
        )
        self._regs = {}
        # 32 bit registers
        for reg_name in (
            "R0",
            "R1",
            "R2",
            "R3",
            "R4",
            "R5",
            "R6",
            "R7",
            "R8",
            "R9",
            "R10",
            "R11",
            "R12",
            "R13",
            "R14",
            "R15",
        ):
            self._regs[reg_name] = Register(32)
        # 64 bit registers
        for reg_name in (
            "D0",
            "D1",
            "D2",
            "D3",
            "D4",
            "D5",
            "D6",
            "D7",
            "D8",
            "D9",
            "D10",
            "D11",
            "D12",
            "D13",
            "D14",
            "D15",
            "D16",
            "D17",
            "D18",
            "D19",
            "D20",
            "D21",
            "D22",
            "D23",
            "D24",
            "D25",
            "D26",
            "D27",
            "D28",
            "D29",
            "D30",
            "D31",
        ):
            self._regs[reg_name] = Register(64)
        # Flags
        self._regs["APSR_N"] = Register(1)
        self._regs["APSR_Z"] = Register(1)
        self._regs["APSR_C"] = Register(1)
        self._regs["APSR_V"] = Register(1)
        self._regs["APSR_GE"] = Register(4)

        # MMU Coprocessor  -- to support MCR/MRC for TLS
        self._regs["P15_C13"] = Register(32)

    def _read_APSR(self):
        def make_apsr_flag(flag_expr, offset):
            """Helper for constructing an expression for the APSR register"""
            return Operators.ITEBV(
                32, flag_expr, BitVecConstant(32, 1 << offset), BitVecConstant(32, 0)
            )

        apsr = 0
        N = self.read("APSR_N")
        Z = self.read("APSR_Z")
        C = self.read("APSR_C")
        V = self.read("APSR_V")

        if any(issymbolic(x) for x in [N, Z, C, V]):
            apsr = (
                make_apsr_flag(N, 31)
                | make_apsr_flag(Z, 30)
                | make_apsr_flag(C, 29)
                | make_apsr_flag(V, 28)
            )
        else:
            if N:
                apsr |= 1 << 31
            if Z:
                apsr |= 1 << 30
            if C:
                apsr |= 1 << 29
            if V:
                apsr |= 1 << 28
        return apsr

    def _write_APSR(self, apsr):
        """Auxiliary function - Writes flags from a full APSR (only 4 msb used)"""
        V = Operators.EXTRACT(apsr, 28, 1)
        C = Operators.EXTRACT(apsr, 29, 1)
        Z = Operators.EXTRACT(apsr, 30, 1)
        N = Operators.EXTRACT(apsr, 31, 1)

        self.write("APSR_V", V)
        self.write("APSR_C", C)
        self.write("APSR_Z", Z)
        self.write("APSR_N", N)

    def read(self, register):
        assert register in self
        if register == "APSR":
            return self._read_APSR()
        register = self._alias(register)
        return self._regs[register].read()

    def write(self, register, value):
        assert register in self
        if register == "APSR":
            return self._write_APSR(value)
        register = self._alias(register)
        self._regs[register].write(value)

    @property
    def all_registers(self):
        return super().all_registers + (
            "R0",
            "R1",
            "R2",
            "R3",
            "R4",
            "R5",
            "R6",
            "R7",
            "R8",
            "R9",
            "R10",
            "R11",
            "R12",
            "R13",
            "R14",
            "R15",
            "D0",
            "D1",
            "D2",
            "D3",
            "D4",
            "D5",
            "D6",
            "D7",
            "D8",
            "D9",
            "D10",
            "D11",
            "D12",
            "D13",
            "D14",
            "D15",
            "D16",
            "D17",
            "D18",
            "D19",
            "D20",
            "D21",
            "D22",
            "D23",
            "D24",
            "D25",
            "D26",
            "D27",
            "D28",
            "D29",
            "D30",
            "D31",
            "APSR",
            "APSR_N",
            "APSR_Z",
            "APSR_C",
            "APSR_V",
            "APSR_GE",
            "P15_C13",
        )

    @property
    def canonical_registers(self):
        return (
            "R0",
            "R1",
            "R2",
            "R3",
            "R4",
            "R5",
            "R6",
            "R7",
            "R8",
            "R9",
            "R10",
            "R11",
            "R12",
            "R13",
            "R14",
            "R15",
            "APSR",
        )


class Armv7LinuxSyscallAbi(SyscallAbi):
    """ARMv7 Linux system call ABI"""

    # EABI standards:
    #  syscall # is in R7
    #  arguments are passed in R0-R6
    #  retval is passed in R0
    def syscall_number(self):
        return self._cpu.R7

    def get_arguments(self):
        for i in range(6):
            yield f"R{i}"

    def write_result(self, result):
        self._cpu.R0 = result


class Armv7CdeclAbi(Abi):
    """ARMv7 Cdecl function call ABI"""

    def get_arguments(self):
        # First four passed via R0-R3, then on stack
        for reg in ("R0", "R1", "R2", "R3"):
            yield reg

        for address in self.values_from(self._cpu.STACK):
            yield address

    def write_result(self, result):
        self._cpu.R0 = result

    def ret(self):
        self._cpu.PC = self._cpu.LR


class Armv7Cpu(Cpu):
    """
    Cpu specialization handling the ARMv7 architecture.

    Note: In this implementation, PC contains address of current
    instruction + 4. However, official spec defines PC to be address of
    current instruction + 8 (section A2.3).
    """

    address_bit_size = ARMV7_CPU_ADDRESS_BIT_SIZE
    max_instr_width = 4
    machine = "armv7"
    arch = cs.CS_ARCH_ARM
    # 'mode' is usually defined here as a class member, but it can change, so
    # it's an instance property.

    def __init__(self, memory):
        self._it_conditional = list()
        self._last_flags = {"C": 0, "V": 0, "N": 0, "Z": 0, "GE": 0}
        self._at_symbolic_conditional = None
        self._mode = cs.CS_MODE_ARM
        super().__init__(Armv7RegisterFile(), memory)

    def __getstate__(self):
        state = super().__getstate__()
        state["_last_flags"] = self._last_flags
        state["at_symbolic_conditional"] = self._at_symbolic_conditional
        state["_it_conditional"] = self._it_conditional
        state["_mode"] = self._mode
        return state

    def __setstate__(self, state):
        self._last_flags = state["_last_flags"]
        self._at_symbolic_conditional = state["at_symbolic_conditional"]
        self._it_conditional = state["_it_conditional"]
        self._mode = state["_mode"]
        super().__setstate__(state)

    @property
    def mode(self):
        return self._mode

    @mode.setter
    def mode(self, new_mode):
        assert new_mode in (cs.CS_MODE_ARM, cs.CS_MODE_THUMB)

        if self._mode != new_mode:
            logger.debug(f'swapping into {"ARM" if new_mode == cs.CS_MODE_ARM else "THUMB"} mode')

        self._mode = new_mode
        self.disasm.disasm.mode = new_mode

    def _set_mode_by_val(self, val):
        new_mode = Operators.ITEBV(
            self.address_bit_size, (val & 0x1) == 0x1, cs.CS_MODE_THUMB, cs.CS_MODE_ARM
        )

        if issymbolic(new_mode):
            from ..state import Concretize

            def set_concrete_mode(state, value):
                state.cpu.mode = value

            raise Concretize(
                "Concretizing ARMv7 mode", expression=new_mode, setstate=set_concrete_mode
            )

        self.mode = new_mode

    def _swap_mode(self):
        """Toggle between ARM and Thumb mode"""
        assert self.mode in (cs.CS_MODE_ARM, cs.CS_MODE_THUMB)
        if self.mode == cs.CS_MODE_ARM:
            self.mode = cs.CS_MODE_THUMB
        else:
            self.mode = cs.CS_MODE_ARM

    # Flags that are the result of arithmetic instructions. Unconditionally
    # set, but conditionally committed.
    #
    # Register file has the actual CPU flags
    def set_flags(self, **flags):
        """
        Note: For any unmodified flags, update _last_flags with the most recent
        committed value. Otherwise, for example, this could happen:

            overflow=0
            instr1 computes overflow=1, updates _last_flags, doesn't commit
            instr2 updates all flags in _last_flags except overflow (overflow remains 1 in _last_flags)
            instr2 commits all in _last_flags
            now overflow=1 even though it should still be 0
        """
        unupdated_flags = self._last_flags.keys() - flags.keys()
        for flag in unupdated_flags:
            flag_name = f"APSR_{flag}"
            self._last_flags[flag] = self.regfile.read(flag_name)
        self._last_flags.update(flags)

    def commit_flags(self):
        # XXX: capstone incorrectly sets .update_flags for adc
        if self.instruction.mnemonic == "adc":
            return
        for flag, val in self._last_flags.items():
            flag_name = f"APSR_{flag}"
            self.regfile.write(flag_name, val)

    def _shift(cpu, value, _type, amount, carry):
        """See Shift() and Shift_C() in the ARM manual"""
        assert cs.arm.ARM_SFT_INVALID < _type <= cs.arm.ARM_SFT_RRX_REG

        # XXX: Capstone should set the value of an RRX shift to 1, which is
        # asserted in the manual, but it sets it to 0, so we have to check
        if _type in (cs.arm.ARM_SFT_RRX, cs.arm.ARM_SFT_RRX_REG) and amount != 1:
            amount = 1

        elif _type in range(cs.arm.ARM_SFT_ASR_REG, cs.arm.ARM_SFT_RRX_REG + 1):
            amount = Operators.EXTRACT(amount, 0, 8)

        if amount == 0:
            return value, carry

        width = cpu.address_bit_size

        if _type in (cs.arm.ARM_SFT_ASR, cs.arm.ARM_SFT_ASR_REG):
            return ASR_C(value, amount, width)
        elif _type in (cs.arm.ARM_SFT_LSL, cs.arm.ARM_SFT_LSL_REG):
            return LSL_C(value, amount, width)
        elif _type in (cs.arm.ARM_SFT_LSR, cs.arm.ARM_SFT_LSR_REG):
            return LSR_C(value, amount, width)
        elif _type in (cs.arm.ARM_SFT_ROR, cs.arm.ARM_SFT_ROR_REG):
            return ROR_C(value, amount, width)
        elif _type in (cs.arm.ARM_SFT_RRX, cs.arm.ARM_SFT_RRX_REG):
            return RRX_C(value, carry, width)

        raise NotImplementedError("Bad shift value")

    # TODO add to abstract cpu, and potentially remove stacksub/add from it?
    def stack_push(self, data, nbytes=None):
        if isinstance(data, int):
            nbytes = nbytes or self.address_bit_size // 8
            self.SP -= nbytes
            self.write_int(self.SP, data, nbytes * 8)
        elif isinstance(data, BitVec):
            self.SP -= data.size // 8
            self.write_int(self.SP, data, data.size)
        elif isinstance(data, str):
            self.SP -= len(data)
            self.write(self.SP, data)
        else:
            raise NotImplementedError("unsupported type for stack push data")
        return self.SP

    def stack_peek(self, nbytes=4):
        return self.read(self.SP, nbytes)

    def stack_pop(self, nbytes=4):
        # TODO is the distinction between load and read really in the op size?
        nbits = nbytes * 8
        if nbits == self.address_bit_size:
            val = self.read_int(self.SP, nbits)
        else:
            val = self.read(self.SP, nbytes)
        self.SP += nbytes
        return val

    def read(self, addr, nbytes):
        return self.read_bytes(addr, nbytes)

    def write(self, addr, data):
        return self.write_bytes(addr, data)

    def set_arm_tls(self, data):
        self.regfile.write("P15_C13", data)

    @staticmethod
    def canonicalize_instruction_name(instr):
        name = instr.insn_name().upper()
        # XXX bypass a capstone bug that incorrectly labels some insns as mov
        if name == "MOV":
            if instr.mnemonic.startswith("lsr"):
                return "LSR"
            elif instr.mnemonic.startswith("lsl"):
                return "LSL"
            elif instr.mnemonic.startswith("asr"):
                return "ASR"
        return OP_NAME_MAP.get(name, name)

    def _wrap_operands(self, operands):
        return [Armv7Operand(self, op) for op in operands]

    def should_commit_flags(cpu):
        # workaround for a capstone bug (issue #980);
        # the bug has been fixed the 'master' and 'next' branches of capstone as of 2017-07-31
        if cpu.instruction.id == cs.arm.ARM_INS_UADD8:
            return True

        return cpu.instruction.update_flags

    def should_execute_conditional(cpu):
        # for the IT instruction, the cc applies to the subsequent instructions,
        # so the IT instruction should be executed regardless of its cc
        if cpu.instruction.id == cs.arm.ARM_INS_IT:
            return True

        # support for the it[x[y[z]]] <op> instructions
        if cpu._it_conditional:
            return cpu._it_conditional.pop(0)

        cc = cpu.instruction.cc
        return cpu._evaluate_conditional(cc)

    def _evaluate_conditional(cpu, cc):
        C = cpu.regfile.read("APSR_C")
        N = cpu.regfile.read("APSR_N")
        V = cpu.regfile.read("APSR_V")
        Z = cpu.regfile.read("APSR_Z")

        if cc == cs.arm.ARM_CC_AL:
            ret = True
        elif cc == cs.arm.ARM_CC_EQ:
            ret = Z
        elif cc == cs.arm.ARM_CC_NE:
            ret = Operators.NOT(Z)
        elif cc == cs.arm.ARM_CC_HS:
            ret = C
        elif cc == cs.arm.ARM_CC_LO:
            ret = Operators.NOT(C)
        elif cc == cs.arm.ARM_CC_MI:
            ret = N
        elif cc == cs.arm.ARM_CC_PL:
            ret = Operators.NOT(N)
        elif cc == cs.arm.ARM_CC_VS:
            ret = V
        elif cc == cs.arm.ARM_CC_VC:
            ret = Operators.NOT(V)
        elif cc == cs.arm.ARM_CC_HI:
            ret = Operators.AND(C, Operators.NOT(Z))
        elif cc == cs.arm.ARM_CC_LS:
            ret = Operators.OR(Operators.NOT(C), Z)
        elif cc == cs.arm.ARM_CC_GE:
            ret = N == V
        elif cc == cs.arm.ARM_CC_LT:
            ret = N != V
        elif cc == cs.arm.ARM_CC_GT:
            ret = Operators.AND(Operators.NOT(Z), N == V)
        elif cc == cs.arm.ARM_CC_LE:
            ret = Operators.OR(Z, N != V)
        else:
            raise NotImplementedError("Bad conditional tag")

        return ret

    @instruction
    def IT(cpu):
        cc = cpu.instruction.cc
        true_case = cpu._evaluate_conditional(cc)
        # this is incredibly hacky--how else does capstone expose this?
        # TODO: find a better way than string parsing the mnemonic -GR, 2017-07-13
        for c in cpu.instruction.mnemonic[1:]:
            if c == "t":
                cpu._it_conditional.append(true_case)
            elif c == "e":
                cpu._it_conditional.append(not true_case)

    @instruction
    def UADD8(cpu, dest, src, op):
        op1 = src.read()
        op2 = op.read()
        sums = list()
        carries = list()
        for i in range(4):
            uo1 = UInt(Operators.ZEXTEND(Operators.EXTRACT(op1, (8 * i), 8), 9), 9)
            uo2 = UInt(Operators.ZEXTEND(Operators.EXTRACT(op2, (8 * i), 8), 9), 9)
            byte = uo1 + uo2
            carry = Operators.EXTRACT(byte, 8, 1)
            sums.append(Operators.EXTRACT(byte, 0, 8))
            carries.append(carry)
        dest.write(Operators.CONCAT(32, *reversed(sums)))
        cpu.set_flags(GE=Operators.CONCAT(4, *reversed(carries)))

    @instruction
    def SEL(cpu, dest, op1, op2):
        op1val = op1.read()
        op2val = op2.read()
        result = list()
        GE = cpu.regfile.read("APSR_GE")
        for i in range(4):
            bit = Operators.EXTRACT(GE, i, 1)
            result.append(
                Operators.ITEBV(
                    8, bit, Operators.EXTRACT(op1val, i * 8, 8), Operators.EXTRACT(op2val, i * 8, 8)
                )
            )
        dest.write(Operators.CONCAT(32, *reversed(result)))

    @instruction(can_take_denormalized_mod_imm=True)
    def MOV(cpu, dest, src):
        """
        Implement the MOV{S} instruction.

        Note: If src operand is PC, temporarily release our logical PC
        view and conform to the spec, which dictates PC = curr instr + 8

        :param Armv7Operand dest: The destination operand; register.
        :param Armv7Operand src: The source operand; register or immediate.
        """
        if cpu.mode == cs.CS_MODE_ARM:
            result, carry_out = src.read(with_carry=True)
            dest.write(result)
            cpu.set_flags(C=carry_out, N=HighBit(result), Z=(result == 0))
        else:
            # thumb mode cannot do wonky things to the operand, so no carry calculation
            result = src.read()
            dest.write(result)
            cpu.set_flags(N=HighBit(result), Z=(result == 0))

    @instruction
    def MOVT(cpu, dest, src):
        """
        MOVT writes imm16 to Rd[31:16]. The write does not affect Rd[15:0].

        :param Armv7Operand dest: The destination operand; register
        :param Armv7Operand src: The source operand; 16-bit immediate
        """
        assert src.type == "immediate"
        imm = src.read()
        low_halfword = dest.read() & Mask(16)
        dest.write((imm << 16) | low_halfword)

    @instruction
    def MRC(cpu, coprocessor, opcode1, dest, coprocessor_reg_n, coprocessor_reg_m, opcode2):
        """
        MRC moves to ARM register from coprocessor.

        :param Armv7Operand coprocessor: The name of the coprocessor; immediate
        :param Armv7Operand opcode1: coprocessor specific opcode; 3-bit immediate
        :param Armv7Operand dest: the destination operand: register
        :param Armv7Operand coprocessor_reg_n: the coprocessor register; immediate
        :param Armv7Operand coprocessor_reg_m: the coprocessor register; immediate
        :param Armv7Operand opcode2: coprocessor specific opcode; 3-bit immediate
        """
        assert coprocessor.type == "coprocessor"
        assert opcode1.type == "immediate"
        assert opcode2.type == "immediate"
        assert dest.type == "register"
        imm_coprocessor = coprocessor.read()
        imm_opcode1 = opcode1.read()
        imm_opcode2 = opcode2.read()
        coprocessor_n_name = coprocessor_reg_n.read()
        coprocessor_m_name = coprocessor_reg_m.read()

        if 15 == imm_coprocessor:  # MMU
            if 0 == imm_opcode1:
                if 13 == coprocessor_n_name:
                    if 3 == imm_opcode2:
                        dest.write(cpu.regfile.read("P15_C13"))
                        return
        raise NotImplementedError(
            "MRC: unimplemented combination of coprocessor, opcode, and coprocessor register"
        )

    @instruction
    def LDRD(cpu, dest1, dest2, src, offset=None):
        """Loads double width data from memory."""
        assert dest1.type == "register"
        assert dest2.type == "register"
        assert src.type == "memory"
        mem1 = cpu.read_int(src.address(), 32)
        mem2 = cpu.read_int(src.address() + 4, 32)
        writeback = cpu._compute_writeback(src, offset)
        dest1.write(mem1)
        dest2.write(mem2)
        cpu._cs_hack_ldr_str_writeback(src, offset, writeback)

    @instruction
    def STRD(cpu, src1, src2, dest, offset=None):
        """Writes the contents of two registers to memory."""
        assert src1.type == "register"
        assert src2.type == "register"
        assert dest.type == "memory"
        val1 = src1.read()
        val2 = src2.read()
        writeback = cpu._compute_writeback(dest, offset)
        cpu.write_int(dest.address(), val1, 32)
        cpu.write_int(dest.address() + 4, val2, 32)
        cpu._cs_hack_ldr_str_writeback(dest, offset, writeback)

    @instruction
    def LDREX(cpu, dest, src, offset=None):
        """
        LDREX loads data from memory.
        * If the physical address has the shared TLB attribute, LDREX
          tags the physical address as exclusive access for the current
          processor, and clears any exclusive access tag for this
          processor for any other physical address.
        * Otherwise, it tags the fact that the executing processor has
          an outstanding tagged physical address.

        :param Armv7Operand dest: the destination register; register
        :param Armv7Operand src: the source operand: register
        """
        # TODO: add lock mechanism to underlying memory --GR, 2017-06-06
        cpu._LDR(dest, src, 32, False, offset)

    @instruction
    def STREX(cpu, status, *args):
        """
        STREX performs a conditional store to memory.
        :param Armv7Operand status: the destination register for the returned status; register
        """
        # TODO: implement conditional return with appropriate status --GR, 2017-06-06
        status.write(0)
        return cpu._STR(cpu.address_bit_size, *args)

    def _UXT(cpu, dest, src, src_width):
        """
        Helper for UXT* family of instructions.

        :param ARMv7Operand dest: the destination register; register
        :param ARMv7Operand dest: the source register; register
        :param int src_width: bits to consider of the src operand
        """
        val = GetNBits(src.read(), src_width)
        word = Operators.ZEXTEND(val, cpu.address_bit_size)
        dest.write(word)

    @instruction
    def UXTB(cpu, dest, src):
        """
        UXTB extracts an 8-bit value from a register, zero-extends
        it to the size of the register, and writes the result to the destination register.

        :param ARMv7Operand dest: the destination register; register
        :param ARMv7Operand dest: the source register; register
        """
        cpu._UXT(dest, src, 8)

    @instruction
    def UXTH(cpu, dest, src):
        """
        UXTH extracts an 16-bit value from a register, zero-extends
        it to the size of the register, and writes the result to the destination register.

        :param ARMv7Operand dest: the destination register; register
        :param ARMv7Operand dest: the source register; register
        """
        cpu._UXT(dest, src, 16)

    @instruction
    def PLD(cpu, addr, offset=None):
        """PLD instructs the cpu that the address at addr might be loaded soon."""

    def _compute_writeback(cpu, operand, offset):
        if offset:
            off = offset.read()
        else:
            off = operand.get_mem_offset()
        wbaddr = operand.get_mem_base_addr() + off
        return wbaddr

    def _cs_hack_ldr_str_writeback(cpu, operand, offset, val):
        # capstone bug doesn't set writeback correctly for postindex reg
        if cpu.instruction.writeback or offset:
            operand.writeback(val)

    def _STR(cpu, width, src, dest, offset=None):
        val = src.read()
        writeback = cpu._compute_writeback(dest, offset)
        cpu.write_int(dest.address(), val, width)
        cpu._cs_hack_ldr_str_writeback(dest, offset, writeback)

    @instruction
    def STR(cpu, *args):
        return cpu._STR(cpu.address_bit_size, *args)

    @instruction
    def STRB(cpu, *args):
        return cpu._STR(8, *args)

    @instruction
    def STRH(cpu, *args):
        return cpu._STR(16, *args)

    def _LDR(cpu, dest, src, width, is_signed, offset):
        mem = cpu.read_int(src.address(), width)
        writeback = cpu._compute_writeback(src, offset)
        if is_signed:
            word = Operators.SEXTEND(mem, width, cpu.address_bit_size)
        else:
            word = Operators.ZEXTEND(mem, cpu.address_bit_size)
        if dest.reg in ("PC", "R15"):
            cpu._set_mode_by_val(word)
            word &= ~0x1
            logger.debug(f"LDR writing 0x{word:x} -> PC")
        dest.write(word)
        cpu._cs_hack_ldr_str_writeback(src, offset, writeback)

    @instruction
    def LDR(cpu, dest, src, offset=None):
        cpu._LDR(dest, src, 32, False, offset)

    @instruction
    def LDRH(cpu, dest, src, offset=None):
        cpu._LDR(dest, src, 16, False, offset)

    @instruction
    def LDRSH(cpu, dest, src, offset=None):
        cpu._LDR(dest, src, 16, True, offset)

    @instruction
    def LDRB(cpu, dest, src, offset=None):
        cpu._LDR(dest, src, 8, False, offset)

    @instruction
    def LDRSB(cpu, dest, src, offset=None):
        cpu._LDR(dest, src, 8, True, offset)

    def _ADD(cpu, _op1, _op2, carry=0):
        W = cpu.address_bit_size
        # masking to 32 because sometimes capstone's op.imm field is negative.
        # this converts it back to unsigned
        _op2 = Operators.ZEXTEND(_op2, W)

        uo1 = UInt(_op1, W * 2)
        uo2 = UInt(_op2, W * 2)
        c = UInt(carry, W * 2)
        unsigned_sum = uo1 + uo2 + c

        so1 = SInt(Operators.SEXTEND(_op1, W, W * 2), W * 2)
        so2 = SInt(Operators.SEXTEND(_op2, W, W * 2), W * 2)
        signed_sum = so1 + so2 + c

        result = GetNBits(unsigned_sum, W)

        carry_out = UInt(result, W * 2) != unsigned_sum
        overflow = SInt(Operators.SEXTEND(result, W, W * 2), W * 2) != signed_sum

        cpu.set_flags(C=carry_out, V=overflow, N=HighBit(result), Z=result == 0)

        return result, carry_out, overflow

    @instruction(can_take_denormalized_mod_imm=True)
    def ADC(cpu, dest, op1, op2=None):
        carry = cpu.regfile.read("APSR_C")
        if op2 is not None:
            result, carry, overflow = cpu._ADD(op1.read(), op2.read(), carry)
        else:
            result, carry, overflow = cpu._ADD(dest.read(), op1.read(), carry)
        dest.write(result)
        return result, carry, overflow

    @instruction(can_take_denormalized_mod_imm=True)
    def ADD(cpu, dest, src, add=None):
        if add is not None:
            result, carry, overflow = cpu._ADD(src.read(), add.read())
        else:
            # support for the thumb mode version of adds <dest>, <immediate>
            result, carry, overflow = cpu._ADD(dest.read(), src.read())
        dest.write(result)
        return result, carry, overflow

    @instruction(can_take_denormalized_mod_imm=True)
    def RSB(cpu, dest, src, add):
        inv_src = GetNBits(~src.read(), cpu.address_bit_size)
        result, carry, overflow = cpu._ADD(inv_src, add.read(), 1)
        dest.write(result)
        return result, carry, overflow

    @instruction(can_take_denormalized_mod_imm=True)
    def RSC(cpu, dest, src, add):
        carry = cpu.regfile.read("APSR_C")
        inv_src = GetNBits(~src.read(), cpu.address_bit_size)
        result, carry, overflow = cpu._ADD(inv_src, add.read(), carry)
        dest.write(result)
        return result, carry, overflow

    @instruction(can_take_denormalized_mod_imm=True)
    def SUB(cpu, dest, src, add=None):
        if add is not None:
            result, carry, overflow = cpu._ADD(src.read(), ~add.read(), 1)
        else:
            # support for the thumb mode version of sub <dest>, <immediate>
            result, carry, overflow = cpu._ADD(dest.read(), ~src.read(), 1)

        dest.write(result)
        return result, carry, overflow

    @instruction(can_take_denormalized_mod_imm=True)
    def SBC(cpu, dest, op1, op2=None):
        carry = cpu.regfile.read("APSR_C")
        if op2 is not None:
            result, carry, overflow = cpu._ADD(op1.read(), ~op2.read(), carry)
        else:
            result, carry, overflow = cpu._ADD(dest.read(), ~op1.read(), carry)
        dest.write(result)
        return result, carry, overflow

    @instruction
    def ADR(cpu, dest, src):
        """
        Address to Register adds an immediate value to the PC value, and writes the result to the destination register.

        :param ARMv7Operand dest: Specifies the destination register.
        :param ARMv7Operand src:
            Specifies the label of an instruction or literal data item whose address is to be loaded into
            <Rd>. The assembler calculates the required value of the offset from the Align(PC,4)
            value of the ADR instruction to this label.
        """
        aligned_pc = (cpu.instruction.address + 4) & 0xFFFFFFFC
        dest.write(aligned_pc + src.read())

    @instruction
    def ADDW(cpu, dest, src, add):
        """
        This instruction adds an immediate value to a register value, and writes the result to the destination register.
        It doesn't update the condition flags.

        :param ARMv7Operand dest: Specifies the destination register. If omitted, this register is the same as src.
        :param ARMv7Operand src:
            Specifies the register that contains the first operand. If the SP is specified for dest, see ADD (SP plus
            immediate). If the PC is specified for dest, see ADR.
        :param ARMv7Operand add:
            Specifies the immediate value to be added to the value obtained from src. The range of allowed values is
            0-4095.
        """
        aligned_pc = (cpu.instruction.address + 4) & 0xFFFFFFFC
        if src.type == "register" and src.reg in ("PC", "R15"):
            src = aligned_pc
        else:
            src = src.read()
        dest.write(src + add.read())

    @instruction
    def SUBW(cpu, dest, src, add):
        """
        This instruction subtracts an immediate value from a register value, and writes the result to the destination
        register. It can optionally update the condition flags based on the result.

        :param ARMv7Operand dest: Specifies the destination register. If omitted, this register is the same as src.
        :param ARMv7Operand src:
            Specifies the register that contains the first operand. If the SP is specified for dest, see ADD (SP plus
            immediate). If the PC is specified for dest, see ADR.
        :param ARMv7Operand add:
            Specifies the immediate value to be added to the value obtained from src. The range of allowed values is
            0-4095.
        """
        aligned_pc = (cpu.instruction.address + 4) & 0xFFFFFFFC
        if src.type == "register" and src.reg in ("PC", "R15"):
            src = aligned_pc
        else:
            src = src.read()
        dest.write(src - add.read())

    @instruction
    def B(cpu, dest):
        cpu.PC = dest.read()

    @instruction
    def BX(cpu, dest):
        dest_val = dest.read()
        cpu._set_mode_by_val(dest_val)
        cpu.PC = dest_val & ~1

    @instruction
    def BLE(cpu, dest):
        cpu.PC = Operators.ITEBV(
            cpu.address_bit_size, cpu.regfile.read("APSR_Z"), dest.read(), cpu.PC
        )

    @instruction
    def CBZ(cpu, op, dest):
        """
        Compare and Branch on Zero compares the value in a register with zero, and conditionally branches forward
        a constant value. It does not affect the condition flags.

        :param ARMv7Operand op: Specifies the register that contains the first operand.
        :param ARMv7Operand dest:
            Specifies the label of the instruction that is to be branched to. The assembler calculates the
            required value of the offset from the PC value of the CBZ instruction to this label, then
            selects an encoding that will set imm32 to that offset. Allowed offsets are even numbers in
            the range 0 to 126.
        """
        cpu.PC = Operators.ITEBV(cpu.address_bit_size, op.read(), cpu.PC, dest.read())

    @instruction
    def CBNZ(cpu, op, dest):
        """
        Compare and Branch on Non-Zero compares the value in a register with zero, and conditionally branches
        forward a constant value. It does not affect the condition flags.

        :param ARMv7Operand op: Specifies the register that contains the first operand.
        :param ARMv7Operand dest:
            Specifies the label of the instruction that is to be branched to. The assembler calculates the
            required value of the offset from the PC value of the CBNZ instruction to this label, then
            selects an encoding that will set imm32 to that offset. Allowed offsets are even numbers in
            the range 0 to 126.
        """
        cpu.PC = Operators.ITEBV(cpu.address_bit_size, op.read(), dest.read(), cpu.PC)

    @instruction
    def BL(cpu, label):
        next_instr_addr = cpu.regfile.read("PC")
        if cpu.mode == cs.CS_MODE_THUMB:
            cpu.regfile.write("LR", next_instr_addr + 1)
        else:
            cpu.regfile.write("LR", next_instr_addr)
        cpu.regfile.write("PC", label.read())

    @instruction
    def BLX(cpu, dest):
        address = cpu.PC
        target = dest.read()
        next_instr_addr = cpu.regfile.read("PC")
        if cpu.mode == cs.CS_MODE_THUMB:
            cpu.regfile.write("LR", next_instr_addr + 1)
        else:
            cpu.regfile.write("LR", next_instr_addr)
        cpu.regfile.write("PC", target & ~1)

        # The `blx <label>` form of this instruction forces a mode swap
        # Otherwise check the lsb of the destination and set the mode
        if dest.type == "immediate":
            logger.debug(f"swapping mode due to BLX at inst 0x{address:x}")
            cpu._swap_mode()
        elif dest.type == "register":
            cpu._set_mode_by_val(dest.read())

    @instruction
    def TBB(cpu, dest):
        """
        Table Branch Byte causes a PC-relative forward branch using a table of single byte offsets. A base register
        provides a pointer to the table, and a second register supplies an index into the table. The branch length is
        twice the value of the byte returned from the table.

        :param ARMv7Operand dest: see below; register
        """
        # Capstone merges the two registers values into one operand, so we need to extract them back

        # Specifies the base register. This contains the address of the table of branch lengths. This
        # register is allowed to be the PC. If it is, the table immediately follows this instruction.
        base_addr = dest.get_mem_base_addr()
        if dest.mem.base in ("PC", "R15"):
            base_addr = cpu.PC

        # Specifies the index register. This contains an integer pointing to a single byte within the
        # table. The offset within the table is the value of the index.
        offset = cpu.read_int(base_addr + dest.get_mem_offset(), 8)
        offset = Operators.ZEXTEND(offset, cpu.address_bit_size)

        cpu.PC += offset << 1

    @instruction
    def TBH(cpu, dest):
        """
        Table Branch Halfword causes a PC-relative forward branch using a table of single halfword offsets. A base
        register provides a pointer to the table, and a second register supplies an index into the table. The branch
        length is twice the value of the halfword returned from the table.

        :param ARMv7Operand dest: see below; register
        """
        # Capstone merges the two registers values into one operand, so we need to extract them back

        # Specifies the base register. This contains the address of the table of branch lengths. This
        # register is allowed to be the PC. If it is, the table immediately follows this instruction.
        base_addr = dest.get_mem_base_addr()
        if dest.mem.base in ("PC", "R15"):
            base_addr = cpu.PC

        # Specifies the index register. This contains an integer pointing to a halfword within the table.
        # The offset within the table is twice the value of the index.
        offset = cpu.read_int(base_addr + dest.get_mem_offset(), 16)
        offset = Operators.ZEXTEND(offset, cpu.address_bit_size)

        cpu.PC += offset << 1

    @instruction(can_take_denormalized_mod_imm=True)
    def CMP(cpu, reg, compare):
        notcmp = ~compare.read() & Mask(cpu.address_bit_size)
        cpu._ADD(reg.read(), notcmp, 1)

    @instruction
    def POP(cpu, *regs):
        for reg in regs:
            val = cpu.stack_pop(cpu.address_bit_size // 8)
            if reg.reg in ("PC", "R15"):
                cpu._set_mode_by_val(val)
                val = val & ~0x1
            reg.write(val)

    @instruction
    def PUSH(cpu, *regs):
        high_to_low_regs = [r.read() for r in regs[::-1]]
        for reg in high_to_low_regs:
            cpu.stack_push(reg)

    @instruction
    def CLZ(cpu, dest, src):

        # Check if the |pos| bit is 1, pos being the offset from the MSB
        value = src.read()
        msb = cpu.address_bit_size - 1
        result = 32

        for pos in range(cpu.address_bit_size):
            cond = Operators.EXTRACT(value, pos, 1) == 1
            result = Operators.ITEBV(cpu.address_bit_size, cond, msb - pos, result)

        dest.write(result)

    @instruction
    def NOP(cpu):
        pass

    @instruction
    def REV(cpu, dest, op):
        opval = op.read()
        _bytes = list()
        for i in range(4):
            _bytes.append(Operators.EXTRACT(opval, i * 8, 8))
        dest.write(Operators.CONCAT(32, *_bytes))

    @instruction
    def SXTH(cpu, dest, op):
        _op = op.read()
        dest.write(Operators.SEXTEND(Operators.EXTRACT(_op, 0, 16), 16, 32))

    def _LDM(cpu, insn_id, base, regs):
        """
        LDM (Load Multiple) loads a non-empty subset, or possibly all, of the general-purpose registers from
        sequential memory locations. It is useful for block loads, stack operations and procedure exit sequences.

        :param int insn_id: should be one of ARM_INS_LDM, ARM_INS_LDMIB, ARM_INS_LDMDA, ARM_INS_LDMDB
        :param Armv7Operand base: Specifies the base register.
        :param list[Armv7Operand] regs:
            Is a list of registers. It specifies the set of registers to be loaded by the LDM instruction.
            The registers are loaded in sequence, the lowest-numbered register from the lowest memory
            address (start_address), through to the highest-numbered register from the highest memory
            address (end_address). If the PC is specified in the register list (opcode bit[15] is set),
            the instruction causes a branch to the address (data) loaded into the PC.

        It's technically UNKNOWN if you writeback to a register you loaded into, but we let it slide.
        """
        if cpu.instruction.usermode:
            raise NotImplementedError("Use of the S bit is not supported")

        increment = insn_id in (cs.arm.ARM_INS_LDM, cs.arm.ARM_INS_LDMIB)
        after = insn_id in (cs.arm.ARM_INS_LDM, cs.arm.ARM_INS_LDMDA)

        address = base.read()

        for reg in regs:
            if not after:
                address += (1 if increment else -1) * (reg.size // 8)

            reg.write(cpu.read_int(address, reg.size))
            if reg.reg in ("PC", "R15"):
                # The general-purpose registers loaded can include the PC. If they do, the word loaded for the PC is
                # treated as an address and a branch occurs to that address. In ARMv5 and above, bit[0] of the loaded
                # value determines whether execution continues after this branch in ARM state or in Thumb state, as
                # though a BX instruction had been executed.
                cpu._set_mode_by_val(cpu.PC)
                cpu.PC = cpu.PC & ~1

            if after:
                address += (1 if increment else -1) * (reg.size // 8)

        if cpu.instruction.writeback:
            base.writeback(address)

    @instruction
    def LDM(cpu, base, *regs):
        cpu._LDM(cs.arm.ARM_INS_LDM, base, regs)

    @instruction
    def LDMIB(cpu, base, *regs):
        cpu._LDM(cs.arm.ARM_INS_LDMIB, base, regs)

    @instruction
    def LDMDA(cpu, base, *regs):
        cpu._LDM(cs.arm.ARM_INS_LDMDA, base, regs)

    @instruction
    def LDMDB(cpu, base, *regs):
        cpu._LDM(cs.arm.ARM_INS_LDMDB, base, regs)

    def _STM(cpu, insn_id, base, regs):
        """
        STM (Store Multiple) stores a non-empty subset (or possibly all) of the general-purpose registers to
        sequential memory locations.

        :param int insn_id: should be one of ARM_INS_STM, ARM_INS_STMIB, ARM_INS_STMDA, ARM_INS_STMDB
        :param Armv7Operand base: Specifies the base register.
        :param list[Armv7Operand] regs:
            Is a list of registers. It specifies the set of registers to be stored by the STM instruction.
            The registers are stored in sequence, the lowest-numbered register to the lowest
            memory address (start_address), through to the highest-numbered register to the
            highest memory address (end_address).
        """
        if cpu.instruction.usermode:
            raise NotImplementedError("Use of the S bit is not supported")

        increment = insn_id in (cs.arm.ARM_INS_STM, cs.arm.ARM_INS_STMIB)
        after = insn_id in (cs.arm.ARM_INS_STM, cs.arm.ARM_INS_STMDA)

        address = base.read()

        for reg in regs:
            if not after:
                address += (1 if increment else -1) * (reg.size // 8)

            cpu.write_int(address, reg.read(), reg.size)

            if after:
                address += (1 if increment else -1) * (reg.size // 8)

        if cpu.instruction.writeback:
            base.writeback(address)

    @instruction
    def STM(cpu, base, *regs):
        cpu._STM(cs.arm.ARM_INS_STM, base, regs)

    @instruction
    def STMIB(cpu, base, *regs):
        cpu._STM(cs.arm.ARM_INS_STMIB, base, regs)

    @instruction
    def STMDA(cpu, base, *regs):
        cpu._STM(cs.arm.ARM_INS_STMDA, base, regs)

    @instruction
    def STMDB(cpu, base, *regs):
        cpu._STM(cs.arm.ARM_INS_STMDB, base, regs)

    def _bitwise_instruction(cpu, operation, dest, op1, op2=None):
        if op2:
            op2_val, carry = op2.read(with_carry=True)
            result = operation(op1.read(), op2_val)
        else:
            # We _do_ use this form, contrary to what LGTM says
            op1_val, carry = op1.read(with_carry=True)  # lgtm [py/call/wrong-arguments]
            result = operation(op1_val)
        if dest is not None:
            dest.write(result)
        cpu.set_flags(C=carry, N=HighBit(result), Z=(result == 0))

    @instruction(can_take_denormalized_mod_imm=True)
    def ORR(cpu, dest, op1, op2=None):
        if op2 is not None:
            cpu._bitwise_instruction(ops.or_, dest, op1, op2)
        else:
            cpu._bitwise_instruction(ops.or_, dest, dest, op1)

    @instruction(can_take_denormalized_mod_imm=True)
    def ORN(cpu, dest, op1, op2=None):
        if op2 is not None:
            cpu._bitwise_instruction(lambda x, y: x | ~y, dest, op1, op2)
        else:
            cpu._bitwise_instruction(lambda x, y: x | ~y, dest, dest, op1)

    @instruction(can_take_denormalized_mod_imm=True)
    def EOR(cpu, dest, op1, op2=None):
        if op2 is not None:
            cpu._bitwise_instruction(ops.xor, dest, op1, op2)
        else:
            cpu._bitwise_instruction(ops.xor, dest, dest, op1)

    @instruction(can_take_denormalized_mod_imm=True)
    def AND(cpu, dest, op1, op2=None):
        if op2 is not None:
            cpu._bitwise_instruction(ops.and_, dest, op1, op2)
        else:
            cpu._bitwise_instruction(ops.and_, dest, dest, op1)

    @instruction(can_take_denormalized_mod_imm=True)
    def TEQ(cpu, op1, op2=None):
        cpu._bitwise_instruction(ops.xor, None, op1, op2)
        cpu.commit_flags()

    @instruction(can_take_denormalized_mod_imm=True)
    def TST(cpu, op1, op2):
        shifted, carry = op2.read(with_carry=True)
        result = op1.read() & shifted
        cpu.set_flags(N=HighBit(result), Z=(result == 0), C=carry)

    @instruction
    def SVC(cpu, op):
        if op.read() != 0:
            logger.warning(f"Bad SVC number: {op.read():08}")
        raise Interruption(0)

    @instruction(can_take_denormalized_mod_imm=True)
    def CMN(cpu, src, add):
        result, carry, overflow = cpu._ADD(src.read(), add.read())
        return result, carry, overflow

    def _SR(cpu, insn_id, dest, op, *rest):
        """
        Notes on Capstone behavior:
        - In ARM mode, _SR reg has `rest`, but _SR imm does not, its baked into `op`.
        - In ARM mode, `lsr r1, r2` will have a `rest[0]`
        - In Thumb mode, `lsr r1, r2` will have an empty `rest`
        - In ARM mode, something like `lsr r1, 3` will not have `rest` and op will be
            the immediate.
        """
        assert insn_id in (cs.arm.ARM_INS_ASR, cs.arm.ARM_INS_LSL, cs.arm.ARM_INS_LSR)

        if insn_id == cs.arm.ARM_INS_ASR:
            if rest and rest[0].type == "immediate":
                srtype = cs.arm.ARM_SFT_ASR
            else:
                srtype = cs.arm.ARM_SFT_ASR_REG
        elif insn_id == cs.arm.ARM_INS_LSL:
            if rest and rest[0].type == "immediate":
                srtype = cs.arm.ARM_SFT_LSL
            else:
                srtype = cs.arm.ARM_SFT_LSL_REG
        elif insn_id == cs.arm.ARM_INS_LSR:
            if rest and rest[0].type == "immediate":
                srtype = cs.arm.ARM_SFT_LSR
            else:
                srtype = cs.arm.ARM_SFT_LSR_REG

        carry = cpu.regfile.read("APSR_C")
        if rest and rest[0].type == "register":
            # FIXME we should make Operand.op private (and not accessible)
            src_reg = cpu.instruction.reg_name(rest[0].op.reg).upper()
            amount = cpu.regfile.read(src_reg)
            result, carry = cpu._shift(op.read(), srtype, amount, carry)
        elif rest and rest[0].type == "immediate":
            amount = rest[0].read()
            result, carry = cpu._shift(op.read(), srtype, amount, carry)
        elif cpu.mode == cs.CS_MODE_THUMB:
            amount = op.read()
            result, carry = cpu._shift(dest.read(), srtype, amount, carry)
        else:
            result, carry = op.read(with_carry=True)
        dest.write(result)

        cpu.set_flags(N=HighBit(result), Z=(result == 0), C=carry)

    @instruction
    def ASR(cpu, dest, op, *rest):
        cpu._SR(cs.arm.ARM_INS_ASR, dest, op, *rest)

    @instruction
    def LSL(cpu, dest, op, *rest):
        cpu._SR(cs.arm.ARM_INS_LSL, dest, op, *rest)

    @instruction
    def LSR(cpu, dest, op, *rest):
        cpu._SR(cs.arm.ARM_INS_LSR, dest, op, *rest)

    @instruction
    def UMULL(cpu, rdlo, rdhi, rn, rm):
        result = UInt(rn.read(), cpu.address_bit_size * 2) * UInt(
            rm.read(), cpu.address_bit_size * 2
        )
        rdhi.write(Operators.EXTRACT(result, cpu.address_bit_size, cpu.address_bit_size))
        rdlo.write(GetNBits(result, cpu.address_bit_size))
        cpu.set_flags(N=Bit(result, 63), Z=(result == 0))

    @instruction
    def MUL(cpu, dest, src1, src2):
        width = cpu.address_bit_size
        op1 = SInt(src1.read(), width)
        op2 = SInt(src2.read(), width)
        result = op1 * op2
        dest.write(result & Mask(width))
        cpu.set_flags(N=HighBit(result), Z=(result == 0))

    @instruction(can_take_denormalized_mod_imm=True)
    def MVN(cpu, dest, op):
        cpu._bitwise_instruction(lambda x: ~x, dest, op)

    @instruction
    def MLA(cpu, dest, op1, op2, addend):
        width = cpu.address_bit_size
        op1_val = SInt(op1.read(), width)
        op2_val = SInt(op2.read(), width)
        add_val = SInt(addend.read(), width)
        result = op1_val * op2_val + add_val

        dest.write(result & Mask(cpu.address_bit_size))
        cpu.set_flags(N=HighBit(result), Z=(result == 0))

    @instruction(can_take_denormalized_mod_imm=True)
    def BIC(cpu, dest, op1, op2=None):
        if op2 is not None:
            result = (op1.read() & ~op2.read()) & Mask(cpu.address_bit_size)
        else:
            result = (dest.read() & ~op1.read()) & Mask(cpu.address_bit_size)
        dest.write(result)
        cpu.set_flags(N=HighBit(result), Z=(result == 0))

    def _VSTM(cpu, address, *regs):
        for reg in regs:
            cpu.write_int(address, reg.read(), reg.size)
            address += reg.size // 8

        return address

    @instruction
    def VSTMIA(cpu, base, *regs):
        updated_address = cpu._VSTM(base.read(), *regs)
        if cpu.instruction.writeback:
            base.writeback(updated_address)

    @instruction
    def VSTMDB(cpu, base, *regs):
        address = base.read() - cpu.address_bit_size // 8 * len(regs)
        updated_address = cpu._VSTM(address, *regs)
        if cpu.instruction.writeback:
            base.writeback(updated_address)

    @instruction
    def VLDMIA(cpu, base, *regs):
        cpu._LDM(cs.arm.ARM_INS_LDM, base, regs)

    @instruction
    def STCL(cpu, *operands):
        pass

    @instruction
    def DMB(cpu, *operands):
        """
        Used by the the __kuser_dmb ARM Linux user-space handler. This is a nop
        under Manticore's memory and execution model.
        """

    @instruction
    def LDCL(cpu, *operands):
        """Occasionally used in glibc (longjmp in ld.so). Nop under our execution model."""

    @instruction
    def UQSUB8(cpu, dest, op1, op2):
        src1 = op1.read()
        src2 = op2.read()
        result = []
        for offset in reversed(range(0, op1.size, 8)):
            byte1 = Operators.EXTRACT(src1, offset, 8)
            byte2 = Operators.EXTRACT(src2, offset, 8)
            byte_diff = byte1 - byte2
            result.append(Operators.ITEBV(8, byte_diff < 0, 0, byte_diff))
        dest.write(Operators.CONCAT(dest.size, *result))



import logging
import types

import elftools
import os
import shlex
import time
from typing import Callable, Optional, Union
import sys
from elftools.elf.elffile import ELFFile
from elftools.elf.sections import SymbolTableSection

from .state import HookCallback, State, TerminateState
from ..core.manticore import ManticoreBase
from ..core.smtlib import ConstraintSet
from ..core.smtlib.solver import SelectedSolver, issymbolic
from ..exceptions import ManticoreError
from ..utils import log, config

logger = logging.getLogger(__name__)

consts = config.get_group("native")
consts.add("stdin_size", default=256, description="Maximum symbolic stdin size")


class Manticore(ManticoreBase):
    def generate_testcase(self, state, message="test"):
        testcase = super().generate_testcase(state, message)
        self._output.save_testcase(state, testcase, message)

    def __init__(self, path_or_state, argv=None, workspace_url=None, policy="random", **kwargs):
        """
        :param path_or_state: Path to binary or a state (object) to begin from.
        :param argv: arguments passed to the binary.
        """
        if isinstance(path_or_state, str):
            if not os.path.isfile(path_or_state):
                raise OSError(f"{path_or_state} is not an existing regular file")
            initial_state = _make_initial_state(path_or_state, argv=argv, **kwargs)
        else:
            initial_state = path_or_state
        super().__init__(initial_state, workspace_url=workspace_url, policy=policy, **kwargs)

        # Move the following into a linux plugin
        self._assertions = {}
        self.trace = None
        self._linux_machine_arch: str  # used when looking up syscall numbers for sys hooks
        # sugar for 'will_execute_instruction"
        self._hooks = {}
        self._after_hooks = {}
        self._sys_hooks = {}
        self._sys_after_hooks = {}
        self._init_hooks = set()

        from ..platforms.linux import Linux

        if isinstance(initial_state.platform, Linux):
            self._linux_machine_arch = initial_state.platform.current.machine

        # self.subscribe('will_generate_testcase', self._generate_testcase_callback)

    ############################################################################
    # Model hooks + callback
    ############################################################################

    def apply_model_hooks(self, path):
        # TODO(yan): Simplify the partial function application

        # Imported straight from __main__.py; this will be re-written once the new
        # event code is in place.
        import importlib
        from manticore import platforms

        with open(path, "r") as fnames:
            for line in fnames.readlines():
                address, cc_name, name = line.strip().split(" ")
                fmodel = platforms
                name_parts = name.split(".")
                importlib.import_module(f".platforms.{name_parts[0]}", "manticore")
                for n in name_parts:
                    fmodel = getattr(fmodel, n)
                assert fmodel != platforms

                def cb_function(state):
                    state.platform.invoke_model(fmodel, prefix_args=(state.platform,))

                self._model_hooks.setdefault(int(address, 0), set()).add(cb_function)
                self._executor.subscribe("will_execute_instruction", self._model_hook_callback)

    def _model_hook_callback(self, state, instruction):
        pc = state.cpu.PC
        if pc not in self._model_hooks:
            return

        for cb in self._model_hooks[pc]:
            cb(state)

    def _generate_testcase_callback(self, state, testcase, message):
        self._output.save_testcase(state, testcase, message)

    @classmethod
    def linux(
        cls,
        path,
        argv=None,
        envp=None,
        entry_symbol=None,
        symbolic_files=None,
        concrete_start="",
        pure_symbolic=False,
        stdin_size=None,
        **kwargs,
    ):
        """
        Constructor for Linux binary analysis.

        :param str path: Path to binary to analyze
        :param argv: Arguments to provide to the binary
        :type argv: list[str]
        :param envp: Environment to provide to the binary
        :type envp: dict[str, str]
        :param entry_symbol: Entry symbol to resolve to start execution
        :type envp: str
        :param symbolic_files: Filenames to mark as having symbolic input
        :type symbolic_files: list[str]
        :param str concrete_start: Concrete stdin to use before symbolic input
        :param int stdin_size: symbolic stdin size to use
        :param kwargs: Forwarded to the Manticore constructor
        :return: Manticore instance, initialized with a Linux State
        :rtype: Manticore
        """
        if stdin_size is None:
            stdin_size = consts.stdin_size

        try:
            return cls(
                _make_linux(
                    path,
                    argv,
                    envp,
                    entry_symbol,
                    symbolic_files,
                    concrete_start,
                    pure_symbolic,
                    stdin_size,
                ),
                **kwargs,
            )
        except elftools.common.exceptions.ELFError:
            raise ManticoreError(f"Invalid binary: {path}")

    @classmethod
    def decree(cls, path, concrete_start="", **kwargs):
        """
        Constructor for Decree binary analysis.

        :param str path: Path to binary to analyze
        :param str concrete_start: Concrete stdin to use before symbolic input
        :param kwargs: Forwarded to the Manticore constructor
        :return: Manticore instance, initialized with a Decree State
        :rtype: Manticore
        """
        try:
            return cls(_make_decree(path, concrete_start), **kwargs)
        except KeyError:  # FIXME(mark) magic parsing for DECREE should raise better error
            raise ManticoreError(f"Invalid binary: {path}")

    @property
    def binary_path(self):
        """
        Assumes that all states refers to a single common program. Might not be
        true in case program calls execve().
        """
        for st in self.all_states:
            return st.platform.program

    ############################################################################
    # Assertion hooks + callback
    ############################################################################

    def load_assertions(self, path):
        with open(path, "r") as f:
            for line in f.readlines():
                pc = int(line.split(" ")[0], 16)
                if pc in self._assertions:
                    logger.debug("Repeated PC in assertions file %s", path)
                self._assertions[pc] = " ".join(line.split(" ")[1:])
                self.subscribe("will_execute_instruction", self._assertions_callback)

    def _assertions_callback(self, state, pc, instruction):
        if pc not in self._assertions:
            return

        from ..core.parser.parser import parse

        program = self._assertions[pc]

        # This will interpret the buffer specification written in INTEL ASM.
        # (It may dereference pointers)
        assertion = parse(program, state.cpu.read_int, state.cpu.read_register)
        if not state.can_be_true(assertion):
            logger.info(str(state.cpu))
            logger.info(
                "Assertion %x -> {%s} does not hold. Aborting state.", state.cpu.PC, program
            )
            raise TerminateState()

        # Everything is good add it.
        state.constraints.add(assertion)

    ###############################
    # Hook Callback Methods
    ###############################

    def init(self, f):
        """
        A decorator used to register a hook function to run before analysis begins. Hook
        function takes one :class:`~manticore.core.state.State` argument.
        """
        self._init_hooks.add(f)
        if self._init_hooks:
            self.subscribe("will_run", self._init_callback)
        return f

    def hook(
        self, pc_or_sys: Optional[Union[int, str]], after: bool = False, syscall: bool = False
    ):
        """
        A decorator used to register a hook function for a given instruction address.
        Equivalent to calling :func:`~add_hook`.

        :param pc_or_sys: Address of instruction, syscall number, or syscall name to remove hook from
        :type pc_or_sys: int or None if `syscall` = False. int, str, or None if `syscall` = True
        :param after: Hook after PC (or after syscall) executes?
        :param syscall: Catch a syscall invocation instead of instruction?
        """

        def decorator(f):
            self.add_hook(pc_or_sys, f, after, None, syscall)
            return f

        return decorator

    def add_hook(
        self,
        pc_or_sys: Optional[Union[int, str]],
        callback: HookCallback,
        after: bool = False,
        state: Optional[State] = None,
        syscall: bool = False,
    ):
        """
        Add a callback to be invoked on executing a program counter (or syscall). Pass `None`
        for `pc_or_sys` to invoke callback on every instruction (or syscall). `callback` should
        be a callable that takes one :class:`~manticore.core.state.State` argument.

        :param pc_or_sys: Address of instruction, syscall number, or syscall name to remove hook from
        :type pc_or_sys: int or None if `syscall` = False. int, str, or None if `syscall` = True
        :param callback: Hook function
        :param after: Hook after PC (or after syscall) executes?
        :param state: Optionally, add hook for this state only, else all states
        :param syscall: Catch a syscall invocation instead of instruction?
        """
        if not (isinstance(pc_or_sys, int) or pc_or_sys is None or syscall):
            raise TypeError(f"pc must be either an int or None, not {pc_or_sys.__class__.__name__}")
        elif not (isinstance(pc_or_sys, (int, str)) or pc_or_sys is None) and syscall:
            raise TypeError(
                f"syscall must be either an int, string, or None, not {pc_or_sys.__class__.__name__}"
            )

        if isinstance(pc_or_sys, str):
            from ..platforms import linux_syscalls

            table = getattr(linux_syscalls, self._linux_machine_arch)
            for index, name in table.items():
                if name == pc_or_sys:
                    pc_or_sys = index
                    break
            if isinstance(pc_or_sys, str):
                logger.warning(
                    f"{pc_or_sys} is not a valid syscall name in architecture {self._linux_machine_arch}. "
                    "Please refer to manticore/platforms/linux_syscalls.py to find the correct name."
                )
                return

        if state is None:
            # add hook to all states
            if not syscall:
                hooks, when, hook_callback = (
                    (self._hooks, "will_execute_instruction", self._hook_callback)
                    if not after
                    else (self._after_hooks, "did_execute_instruction", self._after_hook_callback)
                )
            else:
                hooks, when, hook_callback = (
                    (self._sys_hooks, "will_invoke_syscall", self._sys_hook_callback)
                    if not after
                    else (
                        self._sys_after_hooks,
                        "did_invoke_syscall",
                        self._sys_after_hook_callback,
                    )
                )
            hooks.setdefault(pc_or_sys, set()).add(callback)
            if hooks:
                self.subscribe(when, hook_callback)
        else:
            # only hook for the specified state
            state.add_hook(pc_or_sys, callback, after, syscall)

    def _hook_callback(self, state, pc, instruction):
        "Invoke all registered generic hooks"

        # Ignore symbolic pc.
        # TODO(yan): Should we ask the solver if any of the hooks are possible,
        # and execute those that are?

        if issymbolic(pc):
            return

        # Invoke all pc-specific hooks
        for cb in self._hooks.get(pc, []):
            cb(state)

        # Invoke all pc-agnostic hooks
        for cb in self._hooks.get(None, []):
            cb(state)

    def _after_hook_callback(self, state, last_pc, pc, instruction):
        "Invoke all registered generic hooks"

        # Invoke all pc-specific hooks
        for cb in self._after_hooks.get(last_pc, []):
            cb(state)

        # Invoke all pc-agnostic hooks
        for cb in self._after_hooks.get(None, []):
            cb(state)

    def _sys_hook_callback(self, state, syscall_num):
        "Invoke all registered generic hooks"

        # Invoke all syscall_num-specific hooks
        for cb in self._sys_hooks.get(syscall_num, []):
            cb(state)

        # Invoke all syscall_num-agnostic hooks
        for cb in self._sys_hooks.get(None, []):
            cb(state)

    def _sys_after_hook_callback(self, state, syscall_num):
        "Invoke all registered generic hooks"

        # Invoke all syscall_num-specific hooks
        for cb in self._sys_after_hooks.get(syscall_num, []):
            cb(state)

        # Invoke all syscall_num-agnostic hooks
        for cb in self._sys_after_hooks.get(None, []):
            cb(state)

    def _init_callback(self, ready_states):
        for cb in self._init_hooks:
            # We _should_ only ever have one starting state. Right now we're putting
            # this in a loop for futureproofing.
            for state in ready_states:
                cb(state)

    ###############################
    # Symbol Resolution
    ###############################

    def resolve(self, symbol):
        """
        A helper method used to resolve a symbol name into a memory address when
        injecting hooks for analysis.

        :param symbol: function name to be resolved
        :type symbol: string

        :param line: if more functions present, optional line number can be included
        :type line: int or None
        """

        with open(self.binary_path, "rb") as f:

            elffile = ELFFile(f)

            # iterate over sections and identify symbol table section
            for section in elffile.iter_sections():
                if not isinstance(section, SymbolTableSection):
                    continue

                # get list of symbols by name
                symbols = section.get_symbol_by_name(symbol)
                if not symbols:
                    continue

                # return first indexed memory address for the symbol,
                return symbols[0].entry["st_value"]

            raise ValueError(f"The {self.binary_path} ELFfile does not contain symbol {symbol}")

    def run(self, timeout=None):
        with self.kill_timeout(timeout):
            super().run()

    def finalize(self):
        super().finalize()
        self.save_run_data()

    def save_run_data(self):
        super().save_run_data()


def _make_initial_state(binary_path, **kwargs):
    with open(binary_path, "rb") as f:
        magic = f.read(4)
    if magic == b"\x7fELF":
        # Linux
        state = _make_linux(binary_path, **kwargs)
    elif magic == b"\x7fCGC":
        # Decree
        state = _make_decree(binary_path, **kwargs)
    else:
        raise NotImplementedError(f"Binary {binary_path} not supported.")
    return state


def _make_decree(program, concrete_start="", **kwargs):
    from ..platforms import decree

    constraints = ConstraintSet()
    platform = decree.SDecree(constraints, program)
    initial_state = State(constraints, platform)
    logger.info("Loading program %s", program)

    if concrete_start != "":
        logger.info(f"Starting with concrete input: {concrete_start}")
    platform.input.transmit(concrete_start)
    platform.input.transmit(initial_state.symbolicate_buffer("+" * 14, label="RECEIVE"))
    return initial_state


def _make_linux(
    program,
    argv=None,
    env=None,
    entry_symbol=None,
    symbolic_files=None,
    concrete_start="",
    pure_symbolic=False,
    stdin_size=None,
    *args,
    **kwargs,
) -> State:
    from ..platforms import linux

    env = {} if env is None else env
    argv = [] if argv is None else argv
    env = [f"{k}={v}" for k, v in env.items()]

    if stdin_size is None:
        stdin_size = consts.stdin_size

    logger.info("Loading program %s", program)

    constraints = ConstraintSet()
    platform = linux.SLinux(
        program, argv=argv, envp=env, symbolic_files=symbolic_files, pure_symbolic=pure_symbolic
    )
    if entry_symbol is not None:
        entry_pc = platform._find_symbol(entry_symbol)
        if entry_pc is None:
            logger.error("No symbol for '%s' in %s", entry_symbol, program)
            raise ManticoreError("Symbol not found")
        else:
            logger.info("Found symbol '%s' (%x)", entry_symbol, entry_pc)
            # TODO: use argv as arguments for function
            platform.set_entry(entry_pc)

    initial_state = State(constraints, platform)

    if concrete_start != "":
        logger.info("Starting with concrete input: %s", concrete_start)

    if pure_symbolic:
        logger.warning("[EXPERIMENTAL] Using purely symbolic memory.")

    for i, arg in enumerate(argv):
        argv[i] = initial_state.symbolicate_buffer(arg, label=f"ARGV{i + 1}")

    for i, evar in enumerate(env):
        env[i] = initial_state.symbolicate_buffer(evar, label=f"ENV{i + 1}")

    # If any of the arguments or environment refer to symbolic values, re-
    # initialize the stack
    if any(issymbolic(x) for val in argv + env for x in val):
        platform.setup_stack([program] + argv, env)

    platform.input.write(concrete_start)

    # set stdin input...
    platform.input.write(initial_state.symbolicate_buffer("+" * stdin_size, label="STDIN"))
    return initial_state



import os
import sys
from setuptools import setup, find_packages
from datetime import date

on_rtd = os.environ.get("READTHEDOCS") == "True"


def rtd_dependent_deps():
    # RTD tries to build z3, ooms, and fails to build.
    if on_rtd:
        return native_deps
    else:
        return ["z3-solver"]


# If you update native_deps please update the `REQUIREMENTS_TO_IMPORTS` dict in `utils/install_helper.py`
# (we need to know how to import a given native dependency so we can check if native dependencies are installed)
native_deps = ["capstone==4.0.2", "pyelftools", "unicorn==1.0.2"]

lint_deps = ["black==20.8b1", "mypy==0.790"]

auto_test_deps = ["py-evm"]

# Development dependencies without keystone
dev_noks = (
    native_deps
    + ["coverage", "Sphinx", "pytest==5.3.0", "pytest-xdist==1.30.0", "pytest-cov==2.8.1", "jinja2"]
    + lint_deps
    + auto_test_deps
)

extra_require = {
    "native": native_deps,
    # noks - no keystone
    "dev-noks": dev_noks,
    "dev": native_deps + dev_noks + ["keystone-engine"],
    "redis": ["redis"],
    "lint": lint_deps,
}

this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()


# https://stackoverflow.com/a/4792601 grumble grumble
dev_extension = ""
if "--dev_release" in sys.argv:
    dev_extension = ".dev" + date.today().strftime("%y%m%d")
    sys.argv.remove("--dev_release")

setup(
    name="manticore",
    description="Manticore is a symbolic execution tool for analysis of binaries and smart contracts.",
    long_description_content_type="text/markdown",
    long_description=long_description,
    url="https://github.com/trailofbits/manticore",
    author="Trail of Bits",
    version="0.3.5" + dev_extension,
    packages=find_packages(exclude=["tests", "tests.*"]),
    python_requires=">=3.6",
    install_requires=[
        "pyyaml",
        "protobuf",
        # evm dependencies
        "pysha3",
        "prettytable",
        "ply",
        "rlp",
        "intervaltree",
        "crytic-compile>=0.2.0",
        "wasm",
        "dataclasses; python_version < '3.7'",
        "pyevmasm>=0.2.3",
    ]
    + rtd_dependent_deps(),
    extras_require=extra_require,
    entry_points={
        "console_scripts": [
            "manticore = manticore.__main__:main",
            "manticore-verifier = manticore.ethereum.verifier:main",
        ]
    },
    classifiers=["License :: OSI Approved :: GNU Affero General Public License v3"],
)
